#!/usr/bin/env perl

# $Id: myisam_compression.pl,v 1.17 2003/01/28 21:02:25 mokrejs Exp $

=head1 NAME 

myisam_compression.pl - a simple perl script to (un)compress MyISAM tables

=head1 SYNOPSIS

=head2 myisam_compression.pl --options --dbdir=/data/mysql

=head1 DESCRIPTION

This script can compress and uncompress MyISAM tables as you wish.
The aim of this script was to make the life of MySQL administrators
easier, as the manual procedure for (un)compressing MyISAM tables
can be described by the following commands executed in this order:

  1. FLUSH TABLES;
  2. LOCK TABLE $tablename;
  3. ``compress table''
  4. UNLOCK TABLES;
  5. FLUSH TABLES;

The strong advantage of this program is to selectively include or exclude
some databases or tables which you do or do not want to modify. Hence
we offer --omit-(table|database) and --table and --db arguments.
The program is clever enough to figure out what tables or databases
are present on the system and operate on all of them, if no --db
or --table was specified.

This program *has to* be run locally on the mysql server machine.
It requires full privileges to modify the tables, i.e. should be
run under user `mysql' and requires sql connection to the mysql
daemon with privileges to LOCK/UNLOCK tables and FLUSH table-cache.

=head1 COMPRESSION

The script detemines status of a table using shell command
`myisamchk -dvv $dbdir/$db/$table.MYI'. The possible table formats
are PACKED (i.e. UNCOMPRESSED fields) and COMPRESSED.

If user specified --backup option, it will run compression command with
--backup option. Thus, *.BAK files will be created. You have to
delete them manually! After the *.MYD file is COMPRESSED, we
must regenerate the index file. In this second step the script
executes `myisamchk -rq $dbdir/$db/$table.MYI' command.

=head1 UNCOMPRESSION

In principle uncompressing a table means simply running command
`myisamchk --unpack $dbdir/$db/$table.MYI'. However, this scripts
takes care of flushing table cache and locking table.

As the COMPRESSED table is read-only, we lock the table in READ mode,
not in exclusive (i.e. WRITE) mode.


=head1 ARGUMENTS/OPTIONS

=over 1

=item B< --compress>

Turn this program into the compression mode. Specified databases and
specified tables present in these will be COMPRESSED. The algorithm
will recognize tables already COMPRESSED and will not touch them.

=item B< --db=$dbname>

Specify databases we should (un)compress. This option may be specified
more then once and can be combined with --omit-database=$dbname2
options. --db is not compatible with --dbs-in-file.

If you do specify neither --db nor --dbs-in-file arguments, the program
will create a list of databases from a list of subdirectories present
in $dbdir.

=item B< --dbdir=/data/mysql>

Full path to the directory where databases are stored. That's because
we must change our current working directory to that location.
Only then we perform (un)compression as requested.

=item B< --dbs-in-file=$file>

Filename (might be specified with full path), which contains line by
line list of databases, on which this script should perform (un)compression.
If --dbs-in-file=- is pecified, list of databases is read from STDIN.

=item B< --dry-run>

The program will run as usually, but will not execute any commands which
modify tables.

=item B< --host=$hostname>		[default: 127.0.0.1]

Hostname of the mysql host. Remember that this script can be run
only locally, as we need direct access to the tables. But, as we
need to connect to mysql port and issue LOCK/UNLOCK/FLUSH commands,
we need mysql connection. If you have shut down mysql daemon, then you
obviously don't need this script. ;-)

Note that this program will take the advantage of ~/.my.cnf file.

=item B< --initdb=database>		[default: mysql]

Name of a database to which we connect initially. DBI needs a name
of a database to which it should connect. DBI somehow cannot connect
just to the server, it has to open a database. We won't do anything
with this database, unless is listed in a list of database on which
we should operate.

=item B< --omit-database>

Enumerate a list of databases we should not modify. 

=item B< --omit-table>

Enumerate a list of tables we should not modify. This option is used
to selectively exclude tables, which were specified using --table 
option or, more importantly, if --table was left out, the algorithm would
perform on all tables present in a given database. --omit-table prevents
this program from modifying such tables.

=item B< --password=$password>		[default: ""]

Password to be used during sql connection to LOCK/UNLOCK/FLUSH tables.
Note that this program will take the advantage of ~/.my.cnf file.

=item B< --port=$port>		[default: 3306]

Port number where mysqld is listening to our requests.

=item B< --table=$tablename>

Enumerate a list of tables, which we shouuld (un)compress. If no --table
option is specified, the script will perform (un)compress command on all
tables present in a given database. You can combine --table with
--omit-table options. The script will ignore those tables specified via
--omit-table although they were also specified using --table.

=item B< --tmpdir=$dir>		[default: /var/tmp]

Specify name of a directory, where temporary files created during
compression should be stored.

=item B< --uncompress>

Turn this program into the uncompression mode. Specified databases and
specified tables present in these will be UNCOMPRESSED. The algorithm
will recognize tables already UNCOMPRESSED and will not touch them.

=item B< --user=$user>		[default: ""]

Username to be used during sql connection to LOCK/UNLOCK/FLUSH tables.
Note that this program will take the advantage of ~/.my.cnf file.

=item B< --verbose>

Verbose output. Specifying this option more times increases verbosity.
Maximum is currently 3, although you may specify it more times.


0 If you do not specify --verbose at all, it is set by default to 0.
In such mode, the program will output only Info: about status of tables
before/after compression pluse Error: messages. There're also Warning:
messages displayed, for example when some table is empty, so there's
nothing to compress. 

1 If you specified --verbose only once on a commandline, you will get
default output plus list of tables or databases on which the program
will operate in every step.

2 If you specified --verbose twice, you will get extra messages about
SQL commands executed.

3 Is used more for debugging, you will get extra some output lines
beginning with Debug:. For example, subroutines listing tables
will print some more output.


=head1 EXAMPLES

The examples below show the usage based on the --compress argument.
Imagine --uncompress yourself having same syntax.

myisam_compression.pl --compress --dbdir=/data/mysql --db=db1

Compress all tables present in db1. Read ~/.my.cnf to make SQL connection.

myisam_compression.pl --compress --dbdir=/data/mysql --db=db1 --db=db2 --table=table1

Compress table table1 and table2 in db1 and db2.

myisam_compression.pl --compress --dbdir=/data/mysql --db=db1 --omit-table=table3

Compress all tables present in db1 except table3.

myisam_compression.pl --compress --dbdir=/data/mysql --dbs-in-file=./db-list.txt --omit-database=db2 --omit-table=table3

Compress all tables present in databases listed in db-list.txt, but ignore database db2 and do not compress any table table3, if present.

myisam_compression.pl --compress --dbdir=/data/mysql --dbs-in-file=- --omit-database=db2 --omit-table=table3 < ./db-list.txt

Compress all tables present in databases read from STDIN, but ignore database db2 and do not compress any table table3, if present.

=back

=head1 AUTHOR

Martin Mokrejs <mmokrejs@natur.cuni.cz> http://www.natur.cuni.cz/~mmokrejs/mysql-tools

=cut

my $db;
my @db;
my $dbfile;
my $dbdir;
my $table;
my @table;
my @omitdatabase;
my @omittable;
my @operatedtable;
my @operatedtables;
my $cmd;
my $cmd1;
my $cmd2;
my $cmd3;
my @Cmd;
my @Cmd1;
my @Cmd2;
my @Cmd3;
my $format;
my $verbose=0;
my $user="";
my $host="127.0.0.1";
my $port="3306";
my $password="";
my $initdb="mysql";
my $dbh;
my $compress=0;
my $tmpdir="/var/tmp";
my $uncompress=0;
my $dryrun=0;
my $failed=0;

use Getopt::Long;
use DBI;

GetOptions ('db=s' => \@db, 'dbs-in-file=s' => \$dbfile, 'dbdir=s' => \$dbdir, 'initdb=s' => \$initdb, 'tmpdir=s' => \$tmpdir, 'table=s' => \@table, 'verbose+' => \$verbose, 'host=s' => \$host, 'user=s' => \$user, 'port=s' => \$port, 'password=s' => \$password, 'compress!' => \$compress, 'uncompress!' => \$uncompress, 'omit-table=s' => \@omittable, 'omit-database=s' => \@omitdatabase, 'backup!' => \$backup, 'dry-run!' => \$dryrun);

die "You have to specify --dbdir option to specify where mysql databases reside!\n" if not $dbdir;
if ( ! -x "$dbdir" ) {
  die "Directory $dbdir does not exist.\n";
}

if ( scalar($dbfile) ) {
  if ( "$dbfile" ne "-" ) {
    open (DB,"$dbfile") or die "Cannot open $dbfile: $!\n";
  } elsif ( "$dbfile" eq "-" ) {
    open(DB,"<&STDIN") or die "Cannot open STDIN: $!\n";
  }
  while ( <DB> ) {
    chomp($_);
    push(@db,$_);
  }
  close(DB);
} elsif ( !@db and !$dbfile ) {
  chdir("$dbdir") or die "Cannot chdir() to $dbdir: $!\n";
  foreach ( <*> ) {
    chomp($_);
    push(@db, $_) if ( -d $_);
  }
}

if ( scalar(@omitdatabase) ) {
     @db = RemoveOmitTablesOrDatabases(\@db, \@omitdatabase);
}

chdir("$dbdir") or die "Cannot chdir() to $dbdir: $!\n";

if (ConnectDb("$initdb","$user","$password","$host","$port")) {
  die "Error: Cannot get initial connection to mysql server $host:$port to $initdb database.\n";
}

foreach $db ( @db ) {
  $failed = 0;
  chdir("$dbdir/$db") or ( $failed == 1 );
  if ( $failed == 1 ) {
    warn "Cannot chdir() to $dbdir/$db: $!\n";
    next;
  }
  if (UseDb("$user","$host","$port","$db")) {
    warn "Error: Cannot use database $db on $host:$port. Skipping.\n";
    next;
  }
  $format = "unknown";

  if ( not @table ) {
      @operatedtable = GetTables( "$db", "$db" );
  } else {
    # make sure there's an empty list
      @operatedtable = ();
    #
    # we have to check if the $table is present at all before we put it into @operatedtable array
    #  otherwise let's go for another $table to operate on in this particular database
    #
      my $t = "";
      foreach $t ( @table ) {
	  if ( -r "$t.MYI" ) {
	      print "Debug:$db.$t is present\n" if ( $verbose >= 3 );
	      push(@operatedtable,$t);
	  } else {
	      print "Warning: $db.$t is missing\n";
	  }
      }
  }

  print "Debug: Tables to be omitted from each database: @omittable\n" if $verbose >= 3;
  print "Debug: All tables present in $db or user specified: @operatedtable\n" if $verbose >= 3;

  if ( @omittable and @operatedtable ) {
    @operatedtables = RemoveOmitTablesOrDatabases(\@operatedtable, \@omittable);
  } elsif ( !@operatedtable ) {
    warn "Warning: no tables in database $db\n";
    next;
  } else {
     @operatedtables = @operatedtable;
  }

  print "Info: Final list of tables to operate on in this database $db: @operatedtables\n" if $verbose >= 1;

  foreach $table ( @operatedtables ) {
    $format = "unknown";
    $table =~ s/\..*$//g;
    print "Debug: Looking up format of table $db.$table\n" if $verbose >= 3;
    $format = GetTableStatus($table,$db);

    if ( $format eq "Packed" ) {
      if ( $compress == 1 ) {
        print "Info: $db.$table PACKED ... will be COMPRESSED\n";
        Flush("$host","$port","$db");
        LockTable("$host","$port","$db","$table");
        Flush("$host","$port","$db");
        Compress("$table","$db");
        UnLockTable("$host","$port","$db","$table");
        Flush("$host","$port","$db");
      } else {
        print "Warning: $db.$table PACKED (already uncompressed)\n";
      }
    } elsif ( $format eq "Compressed" ) {
      if ( $uncompress == 1 ) {
        print "Info: $db.$table COMPRESSED ... will be UNCOMPRESSED\n";
        Flush("$host","$port","$db");
        LockTable("$host","$port","$db","$table");
        Flush("$host","$port","$db");
        UnCompress("$table","$db");
        UnLockTable("$host","$port","$db","$table");
        Flush("$host","$port","$db");
      } else {
        print "Info: $db.$table COMPRESSED\n";
      }
    } elsif ( $format eq "Fixed length" ) {
        warn "Warning: [$user\@$host:$port $db.$table] Table is empty, skipping!\n";
    } else {
      die "Error: Ooops! Format is $format.\n";
    }
    print "Info: $db.$table FINISHED OPERATIONS ===================\n" if ( $verbose >= 1 );
  }
  print "Info: $db FINISHED OPERATIONS ===================\n" if ( $verbose >= 1 );
}

$dbh->disconnect;
if ($DBI::errstr) {
   die "Cannot disconnect from mysql server?: $DBI::errstr\n";
} else {
   die "Disconnected from $host:$port.\n";
}


sub ConnectDb {
  my ($db,$user,$password,$host,$port) = @_;
  $dbh = DBI->connect("DBI:mysql:database=$db;mysql_read_default_file=$ENV{HOME}/.my.cnf;host=$host;port=$port","$user","$password",{ RaiseError => 0, AutoCommit => 1, PrintError => 0 });
  if ( $DBI::errstr ) {
    warn "Error: [$user\@$host:$port]: $DBI::errstr\n";
    return 1;
  } else {
    print "Info: Connected to MySQL server $user\@$host:$port. The rest of setting read from $ENV{HOME}/.my.cnf.\nInfo: Will use this connection for FLUSH TABLES, LOCK TABLES WRITE, UNLOCK TABLES\n";
    print "Info: Use --verbose to see those SQL commands\n";
    return 0;
  }
}

sub UseDb {
  my ($user,$host,$port,$db) = @_;
  # we cannot lock READ-ONLY COMPRESSED tables in write mode
  my $statement = "use $db";
  if ( $dryrun == 0 ) {
    $| = 1;
    print "Info: [$user\@$host:$port $db]: Executing $statement ..." if ($verbose >= 2);
    $dbh->do($statement);
    if ( $DBI::errstr ) {
       print " failed.\n" if ($verbose >= 2);
       warn "Error: [$user\@$host:$port $db]: $DBI::errstr\n";
       return 255;
    } else {
       print " succeeded.\n" if ($verbose >= 2);
       print "Info: Using database $db at $user\@$host:$port\n";
       return 0;
    }
  } else {
    print "DRY-RUN: $statement skipped.\n";
    return 0;
  }
}

sub GetTables {
  my ($db) = @_;
  # $dbh->table_info()
  $| = 1;
  print "Info: [$user\@$host:$port $db]: Getting list of tables present ..." if ( $verbose >= 2 );
  my @tables = map { $_ =~ s/.*\.//; $_ } $dbh->tables();
  if ( $DBI::errstr ) {
    print "failed.\n" if ( $verbose >= 2 );
    warn "Error: [$user\@$host:$port $db]: $DBI::errstr\n";
    return;
  } else {
    print " succeeded.\n" if ( $verbose >= 2 );
  }
  foreach ( @tables ) {
    print "Debug: GetTables(): $db.$_\n" if ( $verbose >= 3 );
  }
  return @tables;
}

sub LockTable {
  my ($host,$port,$db,$table) = @_;
  # we cannot lock READ-ONLY COMPRESSED tables in write mode
  my $statement = "LOCK TABLE $table READ;";
  if ( $dryrun == 0 ) {
    $| = 1;
    print "Info: [$user\@$host:$port $db]: Executing $statement ..." if ($verbose >= 2);
    $dbh->do($statement);
    if ( $DBI::errstr ) {
       print " failed.\n" if ($verbose >= 2);
       warn "Error: [$user\@$host:$port $db]: $DBI::errstr\n";
    } else {
       print " succeeded.\n" if ($verbose >= 2);
    }
  } else {
    print "DRY-RUN: $statement skipped.\n";
  }
}
  
sub UnLockTable {
  my ($host,$port,$db,$table) = @_;
  my $statement = "UNLOCK TABLES;";
  if ( $dryrun == 0 ) {
    $| = 1;
    print "Info: [$user\@$host:$port $db]: Executing $statement ..." if ($verbose >= 2);
    $dbh->do($statement) or $failed = 1;
    if ( $DBI::errstr ) {
      print " failed\n" if ($verbose >= 2);
      warn "Error: [$user\@$host:$port $db]: $DBI::errstr\n";
    } else {
      print " succeeded.\n" if ($verbose >= 2);
    }
  } else {
    print "DRY-RUN: $statement skipped.\n";
  }
}

sub Flush {
  my ($host,$port,$db) = @_;
  my $statement = "FLUSH TABLES;";
  if ( $dryrun == 0 ) {
    $| = 1;
    print "Info: [$user\@$host:$port $db]: Executing $statement ..." if ($verbose >= 2);
    $dbh->do($statement);
    if ( ! $DBI::errstr ) {
      print " succeeded.\n" if ($verbose >= 2);
    } else {
      print " failed\n" if ($verbose >= 2);
      warn "Error: [$user\@$host:$port $db]: $DBI::errstr\n";
      $dbh->disconnect;
      die "We shall NOT continue when we cannot flush cache. Disconnected.\n"; 
    }
  } else {
    print "DRY-RUN: $statement skipped.\n";
  }
}

  
sub GetTableStatus { 
  my ($table,$db) = @_;

  if ( -e "/tmp/myisamchk.out.$$" ) {
    unlink("/tmp/myisamchk.out.$$") or die "Cannot unlink /tmp/myisamchk.out.$$: $!\n";
  }

  # We have to catch the output of myisamchk, so that we can read out of it,
  # if the table is PACKED or COMPRESSED.
  #
  open(SAVESTDOUT,">>&STDOUT");
  close(STDOUT);
  open(STDOUT,">>/tmp/myisamchk.out.$$");

  open(SAVESTDERR,">&STDERR");
  close(STDERR);
  open(STDERR,">>/tmp/myisampack.err.$$");

  $cmd3 = "/usr/local/mysql/bin/myisamchk -f -d $dbdir/$db/$table.MYI";
  @Cmd3 = split(/ /,$cmd3);

  system(@Cmd3) and print SAVESTDERR "Error: myisamchk -f -d $dbdir/$db/$table.MYI failed: $!\n" and die;

  close(STDOUT);
  open(STDOUT,">&SAVESTDOUT");
  close(SAVESTDOUT);

  close(STDERR) or print SAVESTDERR "Error: Cannot close STDERR: $!\n";
  open(STDERR,">&SAVESTDERR") or print SAVESTDERR "Cannot dup STDERR from SAVESTDERR: $!\n";
  close(SAVESTDERR) or print STDERR "Cannot close SAVESTDERR: $!\n";
  $| = 1;
  $| = 0;

  $format = "unknown";

  open(OUT,"</tmp/myisamchk.out.$$") or die "Error: Cannot open /tmp/myisamchk.out.$$: $!\n";
  while ( <OUT> ) {
    if ( /^Record\ format:/g ) {
      $format = $_;
      $format =~ s/^Record\ format:[ 	]*(\w+\s*\w*)$/$1/g;
      chomp($format);
    }
  }
  close(OUT);
  unlink("/tmp/myisamchk.out.$$") or die "Error: Cannot unlink /tmp/myisamchk.out.$$: $!\n";

  if ( $format eq "Packed" ) {
    print "Debug: $db.$table is PACKED.\n" if ( $verbose >= 3 );
  } elsif ( $format eq "" ) {
    print STDERR "Error: Ooops, regex pattern wrong?\n";
    $format = "unknown";
  } elsif ( $format eq "unknown" ) {
    die "Error: Ooops! Format unknown.\n";
  } elsif( $format eq "Compressed" ) {
    print "Debug: $db.$table is COMPRESSED.\n" if ( $verbose >= 3 );
  } elsif ( $format eq "Fixed length" ) {
    print STDERR "Debug: $db.$table is empty.\n" if ( $verbose >= 3 );
  } else {
    print STDERR "Error: Ooops! Unexpected table format!\n";
  }
  return ($format);
}


sub Compress {
  my ($table,$db) = @_;
  print "Info: [$dbdir/$db/$table.*]: BEFORE COMPRESSION\n" if ($verbose >= 2);
  my $failed;
  $cmd1 = "/usr/local/mysql/bin/myisampack --backup -f --tmpdir=$tmpdir $dbdir/$db/$table.MYI";
  print "Info: $cmd1\n";
  @Cmd1 = split(/ /,$cmd1);
  if ( $dryrun == 0 ) {
    $failed = 0;
    system(@Cmd1) and ( $failed = 1 );
    if ( $failed == 1 ) {
       warn "Error: [$dbdir/$db/$table.*] COMPRESSION FAILED\n";
       warn "Error: [$dbdir/$db/$table.*] Deleting $tmpdir/$table.TMD and restoring old data file.\n";
       unlink("$tmpdir/$table.TMD") or warn "Error: $tmpdir/$table.TMD was already deleted, good.\n";
       system("mv $table.OLD $table.MYD") or warn "Error: mv $table.OLD $table.MYD failed: $!\nMaybe myisampack restored the original file?\n";
       next;
    } elsif ( ( $backup == 0 ) and ( $failed == 0 ) ) {
       unlink("$table.OLD") or warn "Error: unlink(table.OLD) failed: $!\n";
       unlink("$tmpdir/$table.TMD") or warn "Error: $tmpdir/$table.TMD was already deleted after successfull compression by myisampack itself.\n";
    } else {
       warn  "Error: compression exited with some weird exit status.\n";
    }
  
    print "Info: [$dbdir/$db/$table.*] COMPRESSED BUT BEFORE REPAIRING INDEX\n" if ($verbose >= 2);
    if ( -e "$tmpdir/$table.TMD" ) {
       die "Error: $tmpdir/$table.TMD still exists, but it should NOT! If the myisampack failed previously, the $tmpdir/$table.TMD\nshould be deleted automatically by myisampack and also subsequently by this script, but you should get a warning \"$tmpdir/$table.TMD was already deleted, good.\"\n";
    }

    $failed = 0;
    $cmd2 = "/usr/local/mysql/bin/myisamchk -O sort_buffer_size=256M -O key_buffer_size=512M -O read_buffer_size=8M -O write_buffer_size=8M -rq -f -a -S --tmpdir=$tmpdir $dbdir/$db/$table.MYI";
    print "Info: $cmd2\n";
    @Cmd2 = split(/ /,$cmd2);
    system(@Cmd2) and ( $failed = 1 );
    if ( $failed == 1 ) {
       warn "Error: [$dbdir/$db/$table.*] REAIRING INDEX FAILED. RESTORING OLD INDEX FILE.\n";
       warn "Error: Probably you may want to look into it, maybe look for left $table.TMM file.\n";
       warn "Error: Possibly, uncompress the datafile back.\n";
    }
  }
  print "Info: [$dbdir/$db/$table.*] COMPRESSED WITH REPAIRED INDEX\n" if ($verbose >= 2);
}

sub UnCompress {
  my ($table,$db) = @_;
  $| = 1;
  $| = 0;
  $cmd1 = "/usr/local/mysql/bin/myisamchk --unpack $dbdir/$db/$table.MYI";
  @Cmd1 = split(/ /,$cmd1);

  $cmd2 = "/usr/local/mysql/bin/myisamchk -O sort_buffer_size=256M -O key_buffer_size=512M -O read_buffer_size=8M -O write_buffer_size=8M -o -a -S -R $dbdir/$db/$table.MYI";
  @Cmd2 = split(/ /,$cmd2);

  if ( $dryrun == 0 ) {
    my $failed = 0;
    print "Info: [$dbdir/$db/$table.*] BEFORE UNCOMPRESSION\n" if ($verbose >= 2);
    print "$cmd1\n";
    system(@Cmd1) and ( $failed = 1 );
    if ( $failed == 0 ) {
      print "Info: [$dbdir/$db/$table.*] UNCOMPRESSED\n"  if ($verbose >= 2);
    } else {
      warn "Error: [$dbdir/$db/$table.*] NOT UNCOMPRESSED BECAUSE OF A FAILURE\n";
    }
  }
}


sub RemoveOmitTablesOrDatabases {
  #
  # This subroutine takes as an input two refernces to arrays, containing
  # an array of names. Thesecond array is a list of names which will
  # be removed from the first array. The remaining array contents are returned.
  #

  my ($ar, $omit) = @_;

  if ( not @$ar ) {
      print "Warning: \@operatedtable is empty, what did you specify as --db=?: @$ar\n";
  } elsif ( $verbose >= 3 ) {
      print "Debug: We might work on: @$ar\n";
  }

  if ( not @$omit ) {
      print "Warning: Our bug? \@omittable is empty: @$omit\n";
  } elsif ( $verbose >= 3 ) {
      print "Debug: We should omit @$omit\n";
  }
   
  my %unique;
  my @unique;

  foreach ( @$ar ) {
     chomp($_);
     # we will increase the counter twice, because we have to
     # distinguish if we increased the counter because of an entry
     # in @$ar or @$omit
     #
     $unique{$_}++;
     $unique{$_}++;
  }
  foreach ( @$omit ) {
     chomp($_);
     $unique{$_}++;
  }
  foreach ( sort keys %unique ) {
     if ( $unique{$_} == 1 ) {
	  print "Debug: Table $_ is not present in this database\n" if ( $verbose >= 3 );
          # nothing, it means only table in --omit-table was specified
	  # and is not present in this database at all
     } elsif ( $unique{$_} == 2 ) {
          push (@unique, $_);
     } elsif ( $unique{$_} == 3 ) {
	  print "Debug: Omitting item $_\n" if ( $verbose >= 3 );
     } elsif ( $unique{$_} == 4 ) {
          warn "Warning: One of the arrays contained duplicated entries. Duplicate removed: $_\n" . "Databases/Tables specified: @$ar\n" . "You asked to omit these: @$omit\n";
     }
  }
  @unique = sort @unique;
  return @unique;
}

